

//URSPRUNGSCOPYRIGHT:
/* ====================================================================
 *
 * Copyright (c) 2018 Juerge Liegner  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the author(s) nor the names of any contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR(S) OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 
 // ERWEITERUNGEN:
 ====================================================================
 * edited by: Alois
 * created: 07.01.2020 16:56:53 
 *
 * changes: DTMF via SIP INFO added
 *          Two bell button inputs added for signalling
 *          Control of a door opener relay added
 *
 ====================================================================
 */


#include <ESP8266WiFi.h>
#include <PubSubClient.h>
#include "NTPClient.h"  // https://github.com/taranais/NTPClient
#include <WiFiUdp.h>
#include <ESP8266WebServer.h>
#include "ESP8266HTTPUpdateServer.h"
#include "Sip.h"
#include "config.h"
#include "interval.h"
#include <EEPROM.h>
#include <ArduinoJson.h>


char caSipIn[2048];
char caSipOut[2048];
uint32_t dooropenerDelay = 0;
uint32_t lastReconnectAttempt = 0;
bool startuploop = false;
char ringtime[5][35] = { "Kein Eintrag\0", "Kein Eintrag\0", "Kein Eintrag\0", "Kein Eintrag\0", "Kein Eintrag\0" };
char runtime[30] = "0 Tage, 0 Std., 0 Min.";
uint8_t wifi_config_mode = 0;
char IpString[20];
char topic_state[56];
char topic_ringtimeJSON[56];
char topic_ringtime1[56];
char topic_ringtime2[56];
char topic_ringtime3[56];
char topic_ringtime4[56];
char topic_ringtime5[56];
char topic_dooropener[56];
char topic_infoJSON[56];
char syslog[6][64] = { "---\0", "---\0", "---\0", "---\0", "---\0", "---\0" };


struct config_data {
  uint8_t ip[4];
  uint8_t mask[4];
  uint8_t gw[4];
  uint8_t dns[4];
  char wifissid[25];
  char wifipassword[40];
  char wifi_hostname[25];
  char ntpServerName[25];
  uint8_t dst_aktiv;
  uint8_t mqttIP[4];
  uint16_t mqttport;
  char mqtt_id[32];
  char mqtt_user[10];
  char mqtt_passwd[10];
  uint8_t mqtt_aktiv;
  char sipserver[20];
  char sipuser[20];
  char sippasswd[20];
  uint16_t sipport;
  char sipdialnr[6];
  char sipdialtext[16];
  uint8_t sipdigit;
  uint8_t dooropenertime;
  uint8_t ringtime;
};

struct config_data cfg;
Sip aSip(caSipOut, sizeof(caSipOut));
WiFiUDP ntpUDP;
WiFiClient espClient;
PubSubClient mqtt_client(espClient);
ESP8266WebServer server(80);
ESP8266HTTPUpdateServer httpUpdater;
NTPClient timeClient(ntpUDP);
ADC_MODE(ADC_VCC);

// MQTT Nachrichten empfangen
void callback(char* topic, byte* payload, unsigned int length) {

  payload[length] = '\0';
  String strTopic = String((char*)topic);
  String strPayload = String((char*)payload);
  if (strTopic == topic_dooropener) {
    if (strPayload == "true" && dooropenerDelay == 0) {
      dooropenerDelay = millis();
      digitalWrite(TUEROEFFNERRELAIS_PIN, LOW);  //Türöffner schalten
      systemlog(" Tueroeffner betaetigt per MQTT");
    }
  }
  return;
}

// Mit Mqtt-Broker verbinden
boolean mqttReconnect() {
  if (mqtt_client.connect(cfg.mqtt_id, cfg.mqtt_user, cfg.mqtt_passwd)) {
    // Once connected, publish an announcement...
    mqtt_client.subscribe(topic_dooropener, 1);
    digitalWrite(LED_PIN, HIGH);
    systemlog(" MQTT-Connect");
  }
  return mqtt_client.connected();
}

void publish_ringtime() {
  mqtt_client.publish(topic_ringtime1, ringtime[0]);
  mqtt_client.publish(topic_ringtime2, ringtime[1]);
  mqtt_client.publish(topic_ringtime3, ringtime[2]);
  mqtt_client.publish(topic_ringtime4, ringtime[3]);
  mqtt_client.publish(topic_ringtime5, ringtime[4]);
  StaticJsonDocument<256> ringtime;
  char jsonBuffer_ringtime[256];
  ringtime["1"] = ringtime[0];
  ringtime["2"] = ringtime[1];
  ringtime["3"] = ringtime[2];
  ringtime["4"] = ringtime[3];
  ringtime["5"] = ringtime[4];
  serializeJson(ringtime, jsonBuffer_ringtime);
  mqtt_client.publish(topic_ringtimeJSON, jsonBuffer_ringtime);
  return;
}


void handle_doorbellevent() {
  aSip.Dial(cfg.sipdialnr, cfg.sipdialtext);  // SIP-Anwahl starten
  strcpy(ringtime[4], ringtime[3]);
  strcpy(ringtime[3], ringtime[2]);
  strcpy(ringtime[2], ringtime[1]);
  strcpy(ringtime[1], ringtime[0]);
  strcpy(ringtime[0], timeClient.getFullDateAndTime().c_str());
  if (cfg.mqtt_aktiv == 1) {
    mqtt_client.publish(topic_state, "true");
    publish_ringtime();
  }
  return;
}


String getWifiQuality() {
  int16_t dbm = WiFi.RSSI();
  uint8_t quality_buffer;
  if (dbm <= -100) {
    quality_buffer = 0;
  } else if (dbm >= -50) {
    quality_buffer = 100;
  } else {
    quality_buffer = 2 * (dbm + 100);
  }
  return String(quality_buffer) + " %";
}


void publish_infoJSON() {
  StaticJsonDocument<512> info;
  char jsonBuffer_info[512];
  info["Letzter systemlog"] = syslog[0];
  info["Hostname"] = cfg.wifi_hostname;
  info["Firmware"] = FIRMWARE;
  info["RSSI"] = WiFi.RSSI();
  info["Signal"] = getWifiQuality();
  info["Kanal"] = WiFi.channel();
  info["Resetgrund"] = ESP.getResetReason();
  info["Laufzeit"] = runtime;
  info["VCC"] = (ESP.getVcc() / 1000.0);
  serializeJson(info, jsonBuffer_info);
  mqtt_client.publish(topic_infoJSON, jsonBuffer_info);
  return;
}


void update_runtime() {
  static uint8_t minuten = 0;
  static uint8_t stunden = 0;
  static uint16_t tage = 0;
  minuten += 10;
  if (minuten == 60) {
    stunden++;
    minuten = 0;
  }
  if (stunden == 24) {
    tage++;
    stunden = 0;
  }
  strcpy(runtime, String(tage).c_str());
  strcat(runtime, " Tage, ");
  strcat(runtime, String(stunden).c_str());
  strcat(runtime, " Std., ");
  strcat(runtime, String(minuten).c_str());
  strcat(runtime, " Min.");
  return;
}


void systemlog(String _systemlog) {
  strcpy(syslog[5], syslog[4]);
  strcpy(syslog[4], syslog[3]);
  strcpy(syslog[3], syslog[2]);
  strcpy(syslog[2], syslog[1]);
  strcpy(syslog[1], syslog[0]);
  strcpy(syslog[0], timeClient.getShortDateAndTime().c_str());
  strcat(syslog[0], _systemlog.c_str());
  return;
}

void set_mqttTopics() {
  strcpy(topic_state, cfg.mqtt_id);
  strcat(topic_state, "/Klingelstatus");
  strcpy(topic_ringtimeJSON, cfg.mqtt_id);
  strcat(topic_ringtimeJSON, "/Klingelzeiten/JSON");
  strcpy(topic_ringtime1, cfg.mqtt_id);
  strcat(topic_ringtime1, "/Klingelzeiten/1");
  strcpy(topic_ringtime2, cfg.mqtt_id);
  strcat(topic_ringtime2, "/Klingelzeiten/2");
  strcpy(topic_ringtime3, cfg.mqtt_id);
  strcat(topic_ringtime3, "/Klingelzeiten/3");
  strcpy(topic_ringtime4, cfg.mqtt_id);
  strcat(topic_ringtime4, "/Klingelzeiten/4");
  strcpy(topic_ringtime5, cfg.mqtt_id);
  strcat(topic_ringtime5, "/Klingelzeiten/5");
  strcpy(topic_dooropener, cfg.mqtt_id);
  strcat(topic_dooropener, "/Tueroeffner");
  strcpy(topic_infoJSON, cfg.mqtt_id);
  strcat(topic_infoJSON, "/info");
  return;
}

void setup_wifi() {
  if (wifi_config_mode == 1) {
    WiFi.softAP("config");
  } else {
    IPAddress Ip(cfg.ip[0], cfg.ip[1], cfg.ip[2], cfg.ip[3]);
    IPAddress Gw(cfg.gw[0], cfg.gw[1], cfg.gw[2], cfg.gw[3]);
    IPAddress Mask(cfg.mask[0], cfg.mask[1], cfg.mask[2], cfg.mask[3]);
    IPAddress Dns(cfg.dns[0], cfg.dns[1], cfg.dns[2], cfg.dns[3]);
    WiFi.persistent(false);
    WiFi.mode(WIFI_STA);
    WiFi.hostname(cfg.wifi_hostname);
    WiFi.config(Ip, Gw, Mask, Dns);
    WiFi.begin(cfg.wifissid, cfg.wifipassword);
  }
  return;
}

/*
//WiFi-Setup zur Erstinbetriebnahme ohne AP-Modus nach Änderungen an der EEPROM Struktur
//======================================================================
void setup_wifi() {
    IPAddress Ip(192, 168, 83, 254);
    IPAddress Gw(192, 168, 83, 1);
    IPAddress Mask(255, 255, 255, 0);
    WiFi.setAutoReconnect(true);
    WiFi.persistent(false);    
    WiFi.mode(WIFI_STA);
    WiFi.hostname(cfg.wifi_hostname);    
    WiFi.config(Ip, Gw, Mask);
    WiFi.begin("WLAN14", "strassenverkehrszulassungsordnung");    
     
}*/


void setup() {
  pinMode(KLINGELTASTER_PIN, INPUT_PULLUP);
  pinMode(JUMPER_PIN, INPUT_PULLUP);
  pinMode(LED_PIN, OUTPUT);
  digitalWrite(LED_PIN, HIGH);
  pinMode(TUEROEFFNERRELAIS_PIN, OUTPUT);
  digitalWrite(TUEROEFFNERRELAIS_PIN, HIGH);
  if (digitalRead(JUMPER_PIN) == LOW) {  // Wenn Jumper gesetzt ist, startet der ESP im AP-Modus
    wifi_config_mode = 1;
  }
  EEPROM.begin(sizeof(cfg));
  EEPROM.get(0, cfg);
  EEPROM.end();
  //WIFI konfigurieren und starten
  setup_wifi();
  //initialisiere SIP-Client
  strcpy(IpString, (String(cfg.ip[0]) + "." + String(cfg.ip[1]) + "." + String(cfg.ip[2]) + "." + String(cfg.ip[3])).c_str());
  aSip.Init(cfg.sipserver, cfg.sipport, IpString, cfg.sipport, cfg.sipuser, cfg.sippasswd, cfg.ringtime);
  // Internen Pullup von unbenutzten GPIOS aktivieren
  pinMode(0, INPUT_PULLUP);
  pinMode(1, INPUT_PULLUP);
  pinMode(3, INPUT_PULLUP);
  pinMode(4, INPUT_PULLUP);
  pinMode(14, INPUT_PULLUP);
  pinMode(16, INPUT_PULLUP);
  timeClient.setServerName(cfg.ntpServerName);
  timeClient.begin();
  timeClient.setUpdateInterval(3600000);
  if (cfg.dst_aktiv == 1) {
    timeClient.setTimeOffset(7200);
  } else {
    timeClient.setTimeOffset(3600);
  }
  systemlog("Systemstart");
  systemlog(ESP.getResetReason());
  //MQTT-Client initialisieren
  if (cfg.mqtt_aktiv == 1) {
    set_mqttTopics();
    IPAddress mqttsrv(cfg.mqttIP[0], cfg.mqttIP[1], cfg.mqttIP[2], cfg.mqttIP[3]);
    mqtt_client.setServer(mqttsrv, cfg.mqttport);
    mqtt_client.setCallback(callback);
    mqtt_client.setBufferSize(512);
  }
  setup_webserver();
}

//---------------------------------------------------------------------------------------------------------------------------

void loop() {

  timeClient.update();
  server.handleClient();
  //-------------------------------------------------------------
  int packetSize = aSip.Udp.parsePacket();
  if (packetSize > 0) {
    caSipIn[0] = 0;
    packetSize = aSip.Udp.read(caSipIn, sizeof(caSipIn));
    if (packetSize > 0) {
      caSipIn[packetSize] = 0;
    }
  }
  aSip.HandleUdpPacket((packetSize > 0) ? caSipIn : 0);

  //-------------------------------------------------------------
  // Bei Verbindungsverlust mit MQTT-Broker neu verbinden
  if (cfg.mqtt_aktiv == 1) {
    if (!mqtt_client.connected()) {
      uint32_t now = millis();
      if (now - lastReconnectAttempt > 5000) {
        lastReconnectAttempt = now;
        // Attempt to reconnect
        if (mqttReconnect()) {
          lastReconnectAttempt = 0;
        }
      }
    } else {
      // Client connected
      mqtt_client.loop();
    }
  }
  //-------------------------------------------------------------

  if (startuploop == false && cfg.mqtt_aktiv == 1 && millis() > 15000) {
      mqtt_client.publish(topic_state, "false");
      publish_infoJSON();
      publish_ringtime();
      startuploop = true;
  }
  //------------------------------------------------------------------
  if (!mqtt_client.connected() && wifi_config_mode == 0 && cfg.mqtt_aktiv == 1) {
    INTERVAL(1000UL) {
      digitalWrite(LED_PIN, !digitalRead(LED_PIN));
    }
  }

  if (!mqtt_client.connected() && wifi_config_mode == 1) {
    INTERVAL(300UL) {
      digitalWrite(LED_PIN, !digitalRead(LED_PIN));
    }
  }

  //=================================================================================================================
  // Abfrage der Klingeltaster
  static uint8_t buttonState;
  buttonState = digitalRead(KLINGELTASTER_PIN);  // Status von Eingang KLINGELTASTER_PIN wird an die Variable buttonState übergeben
  static uint8_t bInDial = 0;

  if (buttonState == LOW)  // Wenn Klingeltaster gedrückt...
  {
    digitalWrite(LED_PIN, LOW);  // Setze Ausgang ledPin auf LOW / LED Ein
    if (!bInDial) {
      handle_doorbellevent();                     //
      bInDial = 1;
    }
  }

  if (buttonState == HIGH)  // Wenn Klingeltaster wieder losgelassen....
  {
    if (!aSip.IsBusy()) {
      bInDial = 0;
      digitalWrite(LED_PIN, HIGH);  //Setze Ausgang ledPin auf LOW / LED Aus
    }
  }
  // Abfrage der Klingeltaster Ende
  //=================================================================================================================



  //=================================================================================================================
  // Ansteuerung Türöffner

  //Türöffnerralais schalten per Telefon
  if (aSip.iSignal == cfg.sipdigit && dooropenerDelay == 0) {
    dooropenerDelay = millis();
    digitalWrite(TUEROEFFNERRELAIS_PIN, LOW);
    systemlog(" Tueroeffner betaetigt per Telefon");
  }

  //Türöffnerralais nach ....ms (dooropenertime) wieder abschalten
  if (dooropenerDelay && (millis() - dooropenerDelay) > (cfg.dooropenertime * 1000)) {
    dooropenerDelay = 0;
    aSip.iSignal = 0;
    digitalWrite(TUEROEFFNERRELAIS_PIN, HIGH);  //Setze Ausgang TUEROEFFNERRELAIS_PIN auf HIGH
    if (cfg.mqtt_aktiv == 1) {
      mqtt_client.publish(topic_dooropener, "false");  // Tueröffnertopic wieder auf false setzen
    }
  }
  // Ansteuerung Türöffner Ende
  //=================================================================================================================



  //=================================================================================================================
  // In Intervallen zu erledigende Dinge

  // Alle 10 min Laufzeit aktualisieren und Systeminfos publishen
  INTERVAL(600000UL) {
    update_runtime();
    if (cfg.mqtt_aktiv == 1) {
      publish_infoJSON();
    }
  }

  // Alle 60s WiFi-Verbindung prüfen und bei Bedarf neu verbinden
  INTERVAL(60000UL) {
    if (WiFi.status() != WL_CONNECTED) {
      WiFi.disconnect();
      WiFi.begin(cfg.wifissid, cfg.wifipassword);
      systemlog(" WiFi-Reconnect");
    }
  }
  // In Intervallen zu erledigende Dinge ENDE
  //=================================================================================================================
}
